@@ -7,6 +7,7 @@ pakbase module
 
 from __future__ import print_function
 
+import abc
 import os
 import sys
 import platform
@@ -17,9 +18,68 @@ from numpy.lib.recfunctions import stack_arrays
 
 from .modflow.mfparbc import ModflowParBc as mfparbc
 from .utils import Util2d, Util3d, Transient2d, MfList, check
+from .utils import OptionBlock
 
 
-class Package(object):
+class PackageInterface(object):
+    @property
+    @abc.abstractmethod
+    def name(self):
+        raise NotImplementedError(
+            'must define get_model_dim_arrays in child '
+            'class to use this base class')
+
+    @name.setter
+    @abc.abstractmethod
+    def name(self, name):
+        raise NotImplementedError(
+            'must define get_model_dim_arrays in child '
+            'class to use this base class')
+
+    @property
+    @abc.abstractmethod
+    def parent(self):
+        raise NotImplementedError(
+            'must define get_model_dim_arrays in child '
+            'class to use this base class')
+
+    @parent.setter
+    @abc.abstractmethod
+    def parent(self, name):
+        raise NotImplementedError(
+            'must define get_model_dim_arrays in child '
+            'class to use this base class')
+
+    @property
+    @abc.abstractmethod
+    def package_type(self):
+        raise NotImplementedError(
+            'must define get_model_dim_arrays in child '
+            'class to use this base class')
+
+    @property
+    @abc.abstractmethod
+    def data_list(self):
+        # [data_object, data_object, ...]
+        raise NotImplementedError(
+            'must define get_model_dim_arrays in child '
+            'class to use this base class')
+
+    @abc.abstractmethod
+    def export(self, f, **kwargs):
+        raise NotImplementedError(
+            'must define get_model_dim_arrays in child '
+            'class to use this base class')
+
+    @property
+    @abc.abstractmethod
+    def plotable(self):
+        raise NotImplementedError(
+            'must define plotable in child '
+            'class to use this base class')
+
+
+class Package(PackageInterface):
     """
     Base package class from which most other packages are derived.
 
@@ -32,7 +92,7 @@ class Package(object):
 
         """
         self.parent = parent  # To be able to access the parent modflow object's attributes
-        if (not isinstance(extension, list)):
+        if not isinstance(extension, list):
             extension = [extension]
         self.extension = []
         self.file_name = []
@@ -49,7 +109,7 @@ class Package(object):
         self.fn_path = os.path.join(self.parent.model_ws, self.file_name[0])
         if (not isinstance(name, list)):
             name = [name]
-        self.name = name
+        self._name = name
         if (not isinstance(unit_number, list)):
             unit_number = [unit_number]
         self.unit_number = unit_number
@@ -61,6 +121,7 @@ class Package(object):
         self.allowDuplicates = allowDuplicates
 
         self.acceptable_dtypes = [int, np.float32, str]
+        self.bc_color = "black"
         return
 
     def __repr__(self):
@@ -70,40 +131,42 @@ class Package(object):
             if not (attr in exclude_attributes):
                 if (isinstance(value, list)):
                     if (len(value) == 1):
-                        s = s + ' {0:s} = {1:s}\n'.format(attr, str(value[0]))
+                        s += ' {0:s} = {1:s}\n'.format(attr, str(value[0]))
                     else:
-                        s = s + ' {0:s} (list, items = {1:d}\n'.format(attr,
+                        s += ' {0:s} (list, items = {1:d}\n'.format(attr,
                                                                        len(
                                                                            value))
                 elif (isinstance(value, np.ndarray)):
-                    s = s + ' {0:s} (array, shape = {1:s})\n'.format(attr,
+                    s += ' {0:s} (array, shape = {1:s})\n'.format(attr,
                                                                      value.shape.__str__()[
                                                                      1:-1])
                 else:
-                    s = s + ' {0:s} = {1:s} ({2:s})\n'.format(attr, str(value),
+                    s += ' {0:s} = {1:s} ({2:s})\n'.format(attr, str(value),
                                                               str(type(value))[
                                                               7:-2])
         return s
 
     def __getitem__(self, item):
         if hasattr(self, 'stress_period_data'):
-            # added this check because stress_period_data also used in Oc and Oc88 but is not a MfList
+            # added this check because stress_period_data also used in Oc and
+            # Oc88 but is not a MfList
+            spd = getattr(self, 'stress_period_data')
             if isinstance(item, MfList):
                 if not isinstance(item, list) and not isinstance(item, tuple):
                     assert item in list(
-                        self.stress_period_data.data.keys()), "package.__getitem__() kper " + str(
+                        spd.data.keys()), "package.__getitem__() kper " + str(
                         item) + " not in data.keys()"
-                    return self.stress_period_data[item]
-                else:
-                    if item[1] not in self.dtype.names:
-                        raise Exception(
-                            "package.__getitem(): item \'" + item + "\' not in dtype names " + str(
-                                self.dtype.names))
-                    assert item[0] in list(
-                        self.stress_period_data.data.keys()), "package.__getitem__() kper " + str(
-                        item[0]) + " not in data.keys()"
-                    if self.stress_period_data.vtype[item[0]] == np.recarray:
-                        return self.stress_period_data[item[0]][item[1]]
+                    return spd[item]
+
+                if item[1] not in self.dtype.names:
+                    raise Exception(
+                        "package.__getitem(): item \'" + str(item) +
+                        "\' not in dtype names " + str(self.dtype.names))
+                assert item[0] in list(
+                    spd.data.keys()), "package.__getitem__() kper " + str(
+                    item[0]) + " not in data.keys()"
+                if spd.vtype[item[0]] == np.recarray:
+                    return spd[item[0]][item[1]]
 
     def __setitem__(self, key, value):
         raise NotImplementedError("package.__setitem__() not implemented")
@@ -111,6 +174,8 @@ class Package(object):
     def __setattr__(self, key, value):
         var_dict = vars(self)
         if key in list(var_dict.keys()):
+            if hasattr(self, 'parent'):
+                self.parent._mg_resync = True
             old_value = var_dict[key]
             if isinstance(old_value, Util2d):
                 value = Util2d(self.parent, old_value.shape,
@@ -158,9 +223,49 @@ class Package(object):
 
         super(Package, self).__setattr__(key, value)
 
+    @property
+    def name(self):
+        return self._name
+
+    @name.setter
+    def name(self, name):
+        self._name = name
+
+    @property
+    def parent(self):
+        return self._parent
+
+    @parent.setter
+    def parent(self, parent):
+        self._parent = parent
+
+    @property
+    def package_type(self):
+        if len(self.extension) > 0:
+            return self.extension[0]
+
+    @property
+    def plotable(self):
+        return True
+
+    @property
+    def data_list(self):
+        # return [data_object, data_object, ...]
+        dl = []
+        attrs = dir(self)
+        if 'sr' in attrs:
+            attrs.remove('sr')
+        if 'start_datetime' in attrs:
+            attrs.remove('start_datetime')
+        for attr in attrs:
+            if '__' in attr or 'data_list' in attr:
+                continue
+            dl.append(self.__getattribute__(attr))
+        return dl
+
     def export(self, f, **kwargs):
         from flopy import export
-        return export.utils.package_helper(f, self, **kwargs)
+        return export.utils.package_export(f, self, **kwargs)
 
     @staticmethod
     def add_to_dtype(dtype, field_names, field_types):
@@ -222,22 +327,23 @@ class Package(object):
                         self.name[0] != 'OC':
             spd_inds_valid = True
             chk = check(self, f=f, verbose=verbose, level=level)
-            for per in self.stress_period_data.data.keys():
-                if isinstance(self.stress_period_data.data[per], np.recarray):
-                    spd = self.stress_period_data.data[per]
-                    inds = (spd.k, spd.i, spd.j) if self.parent.structured \
-                        else (spd.node)
+            spd = getattr(self, 'stress_period_data')
+            for per in spd.data.keys():
+                if isinstance(spd.data[per], np.recarray):
+                    spdata = self.stress_period_data.data[per]
+                    inds = (spdata.k, spdata.i, spdata.j) if self.parent.structured \
+                        else (spdata.node)
 
                     # General BC checks
                     # check for valid cell indices
-                    spd_inds_valid = chk._stress_period_data_valid_indices(spd)
+                    spd_inds_valid = chk._stress_period_data_valid_indices(spdata)
 
                     # first check for and list nan values
-                    chk._stress_period_data_nans(spd)
+                    chk._stress_period_data_nans(spdata)
 
                     if spd_inds_valid:
                         # next check for BCs in inactive cells
-                        chk._stress_period_data_inactivecells(spd)
+                        chk._stress_period_data_inactivecells(spdata)
 
                         # More specific BC checks
                         # check elevations in the ghb, drain, and riv packages
@@ -246,8 +352,8 @@ class Package(object):
                             # also checks for nan values
                             elev_name = chk.bc_stage_names[self.name[0]]
                             botms = self.parent.dis.botm.array[inds]
-                            chk.stress_period_data_values(spd, spd[
-                                elev_name] < botms,
+                            chk.stress_period_data_values(spdata,
+                                                          spdata[elev_name] < botms,
                                                           col=elev_name,
                                                           error_name='BC elevation below cell bottom',
                                                           error_type='Error')
@@ -260,7 +366,7 @@ class Package(object):
             chk = check(self, f=f, verbose=verbose, level=level)
             active = chk.get_active()
 
-            # check for confined layers above convertable layers
+            # check for confined layers above convertible layers
             confined = False
             thickstrt = False
             for option in self.options:
@@ -300,7 +406,7 @@ class Package(object):
                            .format(name, mx), 'Warning')
 
             # check for unusually high or low values of hydraulic conductivity
-            if self.layvka.sum() > 0:  # convert vertical anistropy to Kv for checking
+            if self.layvka.sum() > 0:  # convert vertical anisotropy to Kv for checking
                 vka = self.vka.array.copy()
                 for l in range(vka.shape[0]):
                     vka[l] *= self.hk.array[l] if self.layvka.array[
@@ -320,10 +426,9 @@ class Package(object):
                 vkcb = self.vkcb.array.copy()
                 for l in range(self.vkcb.shape[0]):
                     if self.parent.dis.laycbd[l] == 0:
-                        vkcb[l, :,
-                        :] = 1  # assign 1 instead of zero as default value that won't violate checker
+                        # assign 1 instead of zero as default value that won't violate checker
                         # (allows for same structure as other checks)
-
+                        vkcb[l, :, :] = 1
                 chk.values(vkcb, active & (vkcb <= 0),
                            'zero or negative quasi-3D confining bed Kv values',
                            'Error')
@@ -353,7 +458,7 @@ class Package(object):
 
                 # only check specific yield for convertible layers
                 inds = np.array(
-                    [True if l > 0 or l < 0 and 'THICKSRT' in self.options
+                    [True if l > 0 or l < 0 and 'THICKSTRT' in self.options
                      else False for l in self.laytyp])
                 sarrays['sy'] = sarrays['sy'][inds, :, :]
                 active = active[inds, :, :]
@@ -451,130 +556,12 @@ class Package(object):
         >>> ml.dis.plot()
 
         """
+        from flopy.plot import PlotUtilities
 
-        # valid keyword arguments
-        if 'kper' in kwargs:
-            kper = kwargs.pop('kper')
-        else:
-            kper = 0
-
-        if 'filename_base' in kwargs:
-            fileb = kwargs.pop('filename_base')
-        else:
-            fileb = None
-
-        if 'mflay' in kwargs:
-            mflay = kwargs.pop('mflay')
-        else:
-            mflay = None
-
-        if 'file_extension' in kwargs:
-            fext = kwargs.pop('file_extension')
-            fext = fext.replace('.', '')
-        else:
-            fext = 'png'
-
-        if 'key' in kwargs:
-            key = kwargs.pop('key')
-        else:
-            key = None
-
-        if 'initial_fig' in kwargs:
-            ifig = int(kwargs.pop('initial_fig'))
-        else:
-            ifig = 0
-
-        inc = self.parent.nlay
-        if mflay is not None:
-            inc = 1
-
-        axes = []
-        for item, value in self.__dict__.items():
-            caxs = []
-            if isinstance(value, MfList):
-                if self.parent.verbose:
-                    print('plotting {} package MfList instance: {}'.format(
-                        self.name[0], item))
-                if key is None:
-                    names = ['{} location stress period {} layer {}'.format(
-                        self.name[0], kper + 1, k + 1)
-                             for k in range(self.parent.nlay)]
-                    colorbar = False
-                else:
-                    names = ['{} {} data stress period {} layer {}'.format(
-                        self.name[0], key, kper + 1, k + 1)
-                             for k in range(self.parent.nlay)]
-                    colorbar = True
-
-                fignum = list(range(ifig, ifig + inc))
-                ifig = fignum[-1] + 1
-                caxs.append(value.plot(key, names, kper,
-                                       filename_base=fileb,
-                                       file_extension=fext, mflay=mflay,
-                                       fignum=fignum, colorbar=colorbar,
-                                       **kwargs))
-
-            elif isinstance(value, Util3d):
-                if self.parent.verbose:
-                    print('plotting {} package Util3d instance: {}'.format(
-                        self.name[0], item))
-                # fignum = list(range(ifig, ifig + inc))
-                fignum = list(range(ifig, ifig + value.shape[0]))
-                ifig = fignum[-1] + 1
-                caxs.append(
-                    value.plot(filename_base=fileb, file_extension=fext,
-                               mflay=mflay,
-                               fignum=fignum, colorbar=True))
-            elif isinstance(value, Util2d):
-                if len(value.shape) == 2:
-                    if self.parent.verbose:
-                        print('plotting {} package Util2d instance: {}'.format(
-                            self.name[0], item))
-                    fignum = list(range(ifig, ifig + 1))
-                    ifig = fignum[-1] + 1
-                    caxs.append(
-                        value.plot(filename_base=fileb,
-                                   file_extension=fext,
-                                   fignum=fignum, colorbar=True))
-            elif isinstance(value, Transient2d):
-                if self.parent.verbose:
-                    print(
-                        'plotting {} package Transient2d instance: {}'.format(
-                            self.name[0], item))
-                fignum = list(range(ifig, ifig + inc))
-                ifig = fignum[-1] + 1
-                caxs.append(
-                    value.plot(filename_base=fileb, file_extension=fext,
-                               kper=kper,
-                               fignum=fignum, colorbar=True))
-            elif isinstance(value, list):
-                for v in value:
-                    if isinstance(v, Util3d):
-                        if self.parent.verbose:
-                            print(
-                                'plotting {} package Util3d instance: {}'.format(
-                                    self.name[0], item))
-                        fignum = list(range(ifig, ifig + inc))
-                        ifig = fignum[-1] + 1
-                        caxs.append(
-                            v.plot(filename_base=fileb,
-                                   file_extension=fext,
-                                   mflay=mflay,
-                                   fignum=fignum, colorbar=True))
-            else:
-                pass
-
-            # unroll nested lists os axes into a single list of axes
-            if isinstance(caxs, list):
-                for c in caxs:
-                    if isinstance(c, list):
-                        for cc in c:
-                            axes.append(cc)
-                    else:
-                        axes.append(c)
-            else:
-                axes.append(caxs)
+        if not self.plotable:
+            raise TypeError("Package {} is not plotable".format(self.name))
 
+        axes = PlotUtilities._plot_package_helper(self, **kwargs)
         return axes
 
     def to_shapefile(self, filename, **kwargs):
@@ -650,6 +637,13 @@ class Package(object):
             line = f.readline()
             if line[0] != '#':
                 break
+
+        # check for mfnwt version 11 option block
+        nwt_options = None
+        if model.version == "mfnwt" and "options" in line.lower():
+            nwt_options = OptionBlock.load_options(f, pack_type)
+            line = f.readline()
+
         # check for parameters
         nppak = 0
         if "parameter" in line.lower():
@@ -663,6 +657,7 @@ class Package(object):
                     print('   Parameters detected. Number of parameters = ',
                           nppak)
             line = f.readline()
+
         # dataset 2a
         t = line.strip().split()
         ipakcb = 0
@@ -676,38 +671,54 @@ class Package(object):
             it = 2
             while it < len(t):
                 toption = t[it]
-                if toption.lower() is 'noprint':
-                    options.append(toption)
+                if toption.lower() == 'noprint':
+                    options.append(toption.lower())
                 elif 'aux' in toption.lower():
                     options.append(' '.join(t[it:it + 2]))
                     aux_names.append(t[it + 1].lower())
                     it += 1
                 it += 1
 
+        # add auxillary information to nwt options
+        if nwt_options is not None and options:
+            if options[0] == 'noprint':
+                nwt_options.noprint = True
+                if len(options) > 1:
+                    nwt_options.auxillary = options[1:]
+            else:
+                nwt_options.auxillary = options
+
+            options = nwt_options
+
         # set partype
         #  and read phiramp for modflow-nwt well package
         partype = ['cond']
         if 'modflowwel' in str(pack_type).lower():
             partype = ['flux']
 
+        # check for "standard" single line options from mfnwt
         if 'nwt' in model.version.lower() and \
             'flopy.modflow.mfwel.modflowwel'.lower() in str(pack_type).lower():
 
-            specify = False
             ipos = f.tell()
             line = f.readline()
             # test for specify keyword if a NWT well file
             if 'specify' in line.lower():
-                specify = True
-                t = line.strip().split()
-                phiramp = np.float32(t[1])
-                try:
-                    phiramp_unit = np.int32(t[2])
-                except:
-                    phiramp_unit = 2
-                options.append('specify {} {} '.format(phiramp, phiramp_unit))
+                nwt_options = OptionBlock(line.lower().strip(),
+                                          pack_type, block=False)
+                if options:
+                    if options[0] == "noprint":
+                        nwt_options.noprint = True
+                        if len(options) > 1:
+                            nwt_options.auxillary = options[1:]
+
+                    else:
+                        nwt_options.auxillary = options
+
+                options = nwt_options
             else:
                 f.seek(ipos)
+
         elif 'flopy.modflow.mfchd.modflowchd'.lower() in str(
                 pack_type).lower():
             partype = ['shead', 'ehead']
@@ -790,14 +801,14 @@ class Package(object):
                                     current = np.atleast_2d(current).transpose()
                                 #current = np.atleast_2d(np.loadtxt(oc_filename,
                                 #                                   dtype=current.dtype)).transpose()
-                                current = np.core.records.fromarrays(current,dtype=cd)
+                                current = np.core.records.fromarrays(current, dtype=cd)
                             current = current.view(np.recarray)
                         except Exception as e:
                             raise Exception(
                                 "Package.load() error loading open/close file " + oc_filename + \
                                 " :" + str(e))
                         assert current.shape[
-                                   0] == itmp, "Package.load() error: open/close rec array from file " + \
+                                   0] == itmp, "Package.load() error: open/close recarray from file " + \
                                                oc_filename + " shape (" + str(current.shape) + \
                                                ") does not match itmp: {0:d}".format(
                                                    itmp)
